import json
import numpy as np
import pandas as pd
import os
import requests

import matplotlib.pyplot as plt

from collections import Counter
from aimodelshare.aws import run_function_on_lambda, get_aws_client
from aimodelshare.aimsonnx import _get_layer_names, layer_mapping


def get_leaderboard(apiurl, verbose=3, columns=None, submission_type="competition",token=None):
    if token == None:
        if all(["username" in os.environ, 
               "password" in os.environ]):
            pass
        else:
            return print("'get_leaderboard()' unsuccessful. Please provide credentials with set_credentials().")
    else:
        pass
        
    if columns == None: 
        columns = str(columns)

    post_dict = {"y_pred": [],
               "return_eval": "False",
               "return_y": "False",
               "inspect_model": "False",
               "version": "None", 
               "compare_models": "False",
               "version_list": "None",
               "get_leaderboard": "True",
               "submission_type": submission_type,
               "verbose": verbose,
               "columns": columns}
    if token == None:
        token=os.environ.get("AWS_TOKEN")
    else:
       pass
    headers = { 'Content-Type':'application/json', 'authorizationToken': token,} 

    apiurl_eval=apiurl[:-1]+"eval"

    leaderboard_json = requests.post(apiurl_eval,headers=headers,data=json.dumps(post_dict)) 

    leaderboard_pd = pd.DataFrame(json.loads(leaderboard_json.text))

    return leaderboard_pd



def stylize_leaderboard(leaderboard, naming_convention="keras"):

    leaderboard = consolidate_leaderboard(leaderboard, naming_convention=naming_convention)

    # Dropping some columns {{{
    drop_cols = ["timestamp"]
    leaderboard = leaderboard.drop(drop_cols, axis=1)

    #truncate model config info

    if "model_config" in leaderboard.columns:
        leaderboard.model_config = leaderboard.model_config.fillna('None')
        leaderboard.model_config = leaderboard.model_config.map(lambda x: x[0:30]+'...')

    # }}}

    # Setting default properties {{{
    default_props = {"text-align": "center"}

    board = leaderboard.style
    board = board.set_properties(**default_props)
    # }}}

    # infer task type 
    if 'accuracy' in leaderboard.columns.tolist(): 
        task_type = 'classification'
    else: 
        task_type = 'regression'

    # Setting percentage columns' properties {{{
    if task_type == "regression":
        percent_cols = ["mse", 'rmse', 'mae', "r2"]
        percent_colors = ["#f5f8d6", "#c778c8", "#ff4971", "#aadbaa"]

        percent_props = {"color": "#251e1b", "font-size": "12px"}

        for col, color in zip(percent_cols, percent_colors):
            board = board.bar(align="left", color=color, subset=col, vmin=0, vmax=leaderboard[col].max())

        board = board.set_properties(**percent_props, subset=percent_cols)
        board = board.format(lambda x: "{:.2f}".format(x), subset=percent_cols)


    else:
        percent_cols = ["accuracy", "f1_score", "precision", "recall"]
        percent_colors = ["#f5f8d6", "#c778c8", "#ff4971", "#aadbaa"]

        percent_props = {"color": "#251e1b", "font-size": "12px"}

        for col, color in zip(percent_cols, percent_colors):
            board = board.bar(align="left", color=color, subset=col, vmin=0, vmax=1)

        board = board.set_properties(**percent_props, subset=percent_cols)
        board = board.format(lambda x: "{:.2f}%".format(x * 100), subset=percent_cols)
    # }}}

    return board


def consolidate_leaderboard(data, naming_convention="keras"):

  for i in data: 

    i = i.replace('_layers', '')
    i = i.replace('_act', '')

    if 'maxpooling2d_layers' in data.columns and 'maxpool2d_layers' in data.columns:
      matched_cols = ['maxpooling2d_layers', 'maxpool2d_layers']
      sum_cols = data[matched_cols].sum(axis=1)
      data[matched_cols[1]] = sum_cols
      data = data.drop(matched_cols[0], axis=1)

    if naming_convention == 'keras':
      mapping = layer_mapping('torch_to_keras')
      mapping_inverse = layer_mapping('keras_to_torch')
    elif naming_convention == 'pytorch':
      mapping = layer_mapping('keras_to_torch')
      mapping_inverse = layer_mapping('torch_to_keras')

    mapping_lower = {i.lower(): mapping[i].lower() for i in mapping if i is not None and mapping[i] is not None}
    mapping_inverse_lower = {i.lower(): mapping_inverse[i].lower() for i in mapping_inverse if i is not None and mapping_inverse[i] is not None}

    try:

      if i in mapping_lower.keys():

        if not i == mapping_lower.get(i):
          
          matched_cols = [i+"_layers", mapping_lower.get(i)+"_layers"]

          sum_cols = data[matched_cols].sum(axis=1)

          data[matched_cols[1]] = sum_cols

          data = data.drop(matched_cols[0], axis=1)

    except:
      pass

    try:
      if i in mapping_inverse_lower.keys() and naming_convention == 'keras':

        if not i == mapping_inverse_lower.get(i):
          
          matched_cols = [i+"_layers", mapping_inverse_lower.get(i)+"_layers"]

          sum_cols = data[matched_cols].sum(axis=1)

          data[matched_cols[0]] = sum_cols

          data = data.drop(matched_cols[1], axis=1)

    except:
      pass
    
  return data



__all__ = [get_leaderboard,
    stylize_leaderboard]



"""
Moral Compass Integration Helpers for Activities 7, 8, and 9.

This module provides helper functions for integrating the Moral Compass scoring system
into Ethics/Game apps, including:
- ChallengeManager initialization and management
- Debounced server synchronization
- Team aggregation logic
- Leaderboard generation with caching

Design Rationale:
- Client-side only scoring combination logic (server stores single moralCompassScore)
- Debounce prevents excessive API calls while providing responsive UI
- Team synthetic users (prefix: team:) enable team leaderboards
- Local preview fallback ensures graceful degradation when debounced or offline

Server Constraints:
- Only existing API endpoints available (no custom metadata fields)
- All combination logic handled client-side
- Primary metric stored as moralCompassScore in server
"""

import os
import time
import logging
from typing import Optional, Dict, Any, List, Tuple
from datetime import datetime, timedelta

logger = logging.getLogger("aimodelshare.moral_compass.apps")


# ============================================================================
# Environment Configuration
# ============================================================================

def get_env_config() -> Dict[str, Any]:
    """
    Get environment configuration for Moral Compass integration.
    
    Returns:
        Dictionary with configuration values
    """
    return {
        # Debounce settings
        'DEBOUNCE_SECONDS': int(os.getenv('MC_DEBOUNCE_SECONDS', '5')),
        
        # Scoring mode: 'product' or 'sum'
        'SCORING_MODE': os.getenv('MC_SCORING_MODE', 'product'),
        
        # Weights for sum mode
        'WEIGHT_ACCURACY': float(os.getenv('MC_WEIGHT_ACC', '0.6')),
        'WEIGHT_MORAL': float(os.getenv('MC_WEIGHT_MORAL', '0.4')),
        
        # Normalization settings
        'ACCURACY_FLOOR': float(os.getenv('MC_ACCURACY_FLOOR', '0.0')),
        'MAX_MORAL_POINTS': int(os.getenv('MAX_MORAL_POINTS', '1000')),
        
        # Cache TTL for leaderboard
        'CACHE_TTL_SECONDS': int(os.getenv('MC_CACHE_TTL', '30')),
    }


# ============================================================================
# Debounce State Management
# ============================================================================

# Global state for debounce tracking
_last_sync_times: Dict[str, float] = {}


def should_sync(username: str, override: bool = False) -> bool:
    """
    Check if sync should proceed based on debounce logic.
    
    Args:
        username: The username to check
        override: If True, bypass debounce check (for Force Sync)
        
    Returns:
        True if sync should proceed, False if debounced
    """
    if override:
        return True
    
    config = get_env_config()
    debounce_seconds = config['DEBOUNCE_SECONDS']
    
    last_sync = _last_sync_times.get(username, 0)
    current_time = time.time()
    
    return (current_time - last_sync) >= debounce_seconds


def mark_synced(username: str) -> None:
    """
    Mark a username as having been synced.
    
    Args:
        username: The username that was synced
    """
    _last_sync_times[username] = time.time()


# ============================================================================
# ChallengeManager Initialization
# ============================================================================

def get_challenge_manager(username: str, table_id: Optional[str] = None) -> Optional['ChallengeManager']:
    """
    Get or create a ChallengeManager for a user.
    
    Args:
        username: The username
        table_id: Optional table ID (auto-derived if not provided)
        
    Returns:
        ChallengeManager instance, or None if user not signed in
        
    Note:
        Requires aimodelshare.moral_compass.challenge.ChallengeManager
    """
    if not username or username.lower() == 'guest':
        logger.debug("Cannot create ChallengeManager for guest user")
        return None
    
    try:
        from aimodelshare.moral_compass.challenge import ChallengeManager
        from aimodelshare.moral_compass.api_client import MoralcompassApiClient
        
        # Auto-derive table_id if not provided
        if not table_id:
            table_id = _derive_table_id()
        
        # Create API client and ChallengeManager
        api_client = MoralcompassApiClient()
        cm = ChallengeManager(
            table_id=table_id,
            username=username,
            api_client=api_client
        )
        
        logger.info(f"Created ChallengeManager for user={username}, table={table_id}")
        return cm
        
    except Exception as e:
        logger.error(f"Failed to create ChallengeManager: {e}")
        return None


def _derive_table_id() -> str:
    """
    Auto-derive table ID from environment or use default.
    
    Returns:
        Table ID string
    """
    # Check for explicit table ID
    table_id = os.getenv('MORAL_COMPASS_TABLE_ID')
    if table_id:
        return table_id
    
    # Try to derive from playground URL
    playground_url = os.getenv('PLAYGROUND_URL')
    if playground_url:
        # Extract playground ID and append -mc suffix
        from urllib.parse import urlparse
        parsed = urlparse(playground_url)
        path_parts = [p for p in parsed.path.split('/') if p]
        
        for i, part in enumerate(path_parts):
            if part.lower() in ['playground', 'playgrounds']:
                if i + 1 < len(path_parts):
                    playground_id = path_parts[i + 1]
                    return f"{playground_id}-mc"
        
        # Fallback to last path component
        if path_parts:
            return f"{path_parts[-1]}-mc"
    
    # Default fallback
    return "justice-equity-challenge-mc"


# ============================================================================
# Scoring Logic
# ============================================================================

def compute_combined_score(accuracy: float, moral_points: int, 
                           config: Optional[Dict[str, Any]] = None) -> float:
    """
    Compute combined ethical + accuracy score.
    
    Args:
        accuracy: Accuracy value (0.0 to 1.0)
        moral_points: Raw moral compass points
        config: Optional config dict (uses env defaults if None)
        
    Returns:
        Combined score as float
        
    Note:
        All combination logic is client-side. Server receives only the
        final combined score as the primary metric (moralCompassScore).
    """
    if config is None:
        config = get_env_config()
    
    # Apply accuracy floor
    accuracy_floor = config['ACCURACY_FLOOR']
    accuracy = max(accuracy, accuracy_floor)
    
    # Normalize moral points (0 to 1)
    max_moral = config['MAX_MORAL_POINTS']
    moral_normalized = min(moral_points / max_moral, 1.0) if max_moral > 0 else 0.0
    
    # Compute combined score based on mode
    scoring_mode = config['SCORING_MODE']
    
    if scoring_mode == 'product':
        # Product mode: accuracy * moral_normalized
        combined = accuracy * moral_normalized
    elif scoring_mode == 'sum':
        # Weighted sum mode
        weight_acc = config['WEIGHT_ACCURACY']
        weight_moral = config['WEIGHT_MORAL']
        combined = (weight_acc * accuracy) + (weight_moral * moral_normalized)
    else:
        logger.warning(f"Unknown scoring mode '{scoring_mode}', defaulting to product")
        combined = accuracy * moral_normalized
    
    logger.debug(
        f"Combined score: accuracy={accuracy:.4f}, moral_points={moral_points}, "
        f"moral_norm={moral_normalized:.4f}, mode={scoring_mode}, result={combined:.4f}"
    )
    
    return combined


# ============================================================================
# User Sync
# ============================================================================

def sync_user_moral_state(
    cm: 'ChallengeManager',
    moral_points: int,
    accuracy: Optional[float] = None,
    override: bool = False
) -> Dict[str, Any]:
    """
    Sync user's moral state to server with debounce.
    
    Args:
        cm: ChallengeManager instance
        moral_points: Current moral compass points for this activity
        accuracy: Optional accuracy value (fetched from playground if None)
        override: If True, bypass debounce (for Force Sync button)
        
    Returns:
        Dictionary with sync result:
        - 'synced': bool (True if actually synced, False if debounced)
        - 'status': str ('synced', 'debounced', 'error')
        - 'server_score': float (if synced)
        - 'local_preview': float (always present)
        - 'message': str (user-facing message)
        
    Design Note:
        - Seeds ChallengeManager with playground accuracy if not provided
        - Computes combined score (accuracy * moral_normalized) client-side
        - Stores combined score as primary metric on server
        - Respects debounce unless override=True
    """
    username = cm.username
    
    # Check debounce
    if not should_sync(username, override=override):
        local_preview = compute_combined_score(
            accuracy or 0.7,  # Default accuracy for preview
            moral_points
        )
        return {
            'synced': False,
            'status': 'debounced',
            'local_preview': local_preview,
            'message': f'Sync pending (debounced). Local preview: {local_preview:.4f}'
        }
    
    try:
        # Fetch accuracy from playground if not provided
        if accuracy is None:
            accuracy = _fetch_playground_accuracy(username)
        
        # Compute combined score
        combined_score = compute_combined_score(accuracy, moral_points)
        
        # Update ChallengeManager metrics
        cm.set_metric('accuracy', accuracy, primary=False)
        cm.set_metric('moral_points', moral_points, primary=False)
        cm.set_metric('combined_score', combined_score, primary=True)
        
        # Sync to server
        response = cm.sync()
        
        # Mark as synced
        mark_synced(username)
        
        server_score = response.get('moralCompassScore', combined_score)
        
        logger.info(
            f"User sync successful: username={username}, moral_points={moral_points}, "
            f"accuracy={accuracy:.4f}, combined={combined_score:.4f}, "
            f"server_score={server_score:.4f}"
        )
        
        return {
            'synced': True,
            'status': 'synced',
            'server_score': server_score,
            'local_preview': combined_score,
            'message': f'✓ Synced! Server score: {server_score:.4f}'
        }
        
    except Exception as e:
        logger.error(f"User sync failed for {username}: {e}")
        local_preview = compute_combined_score(accuracy or 0.7, moral_points)
        return {
            'synced': False,
            'status': 'error',
            'local_preview': local_preview,
            'error': str(e),
            'message': f'⚠️ Sync error. Local preview: {local_preview:.4f}'
        }


def _fetch_playground_accuracy(username: str) -> float:
    """
    Fetch user's accuracy from playground leaderboard.
    
    Args:
        username: The username
        
    Returns:
        Accuracy value (0.0 to 1.0), defaults to 0.7 if not found
        
    Note:
        Uses playground.get_leaderboard() to fetch accuracy data
    """
    try:
        from aimodelshare.playground import Competition
        
        playground_url = os.getenv('PLAYGROUND_URL', 
                                   'https://cf3wdpkg0d.execute-api.us-east-1.amazonaws.com/prod/m')
        
        playground = Competition(playground_url)
        leaderboard = playground.get_leaderboard()
        
        # Find user's entry
        for entry in leaderboard:
            if entry.get('username') == username or entry.get('user') == username:
                # Get accuracy (might be stored as 'accuracy', 'score', or 'test_accuracy')
                accuracy = (
                    entry.get('accuracy') or 
                    entry.get('test_accuracy') or 
                    entry.get('score', 0.7)
                )
                logger.debug(f"Fetched accuracy for {username}: {accuracy}")
                return float(accuracy)
        
        logger.warning(f"User {username} not found in leaderboard, using default 0.7")
        return 0.7
        
    except Exception as e:
        logger.error(f"Failed to fetch playground accuracy: {e}")
        return 0.7


# ============================================================================
# Team Sync
# ============================================================================

def sync_team_state(team_name: str, table_id: Optional[str] = None) -> Dict[str, Any]:
    """
    Sync team aggregated state to server.
    
    Args:
        team_name: The team name
        table_id: Optional table ID (auto-derived if not provided)
        
    Returns:
        Dictionary with sync result (same structure as sync_user_moral_state)
        
    Design Note:
        - Aggregates member accuracy from playground.get_leaderboard()
        - Aggregates member moral scores from moral_compass.list_users()
        - Computes team combined score (avg_accuracy * avg_moral_norm)
        - Persists as synthetic user with username = 'team:<TeamName>'
    """
    if not team_name:
        return {
            'synced': False,
            'status': 'error',
            'message': 'No team name provided'
        }
    
    try:
        # Auto-derive table_id if not provided
        if not table_id:
            table_id = _derive_table_id()
        
        # Get team members and their data
        team_data = _aggregate_team_data(team_name, table_id)
        
        if not team_data['members']:
            logger.warning(f"No team members found for team '{team_name}'")
            return {
                'synced': False,
                'status': 'error',
                'message': f'No members found for team {team_name}'
            }
        
        # Compute team combined score
        avg_accuracy = team_data['avg_accuracy']
        avg_moral_points = team_data['avg_moral_points']
        
        combined_score = compute_combined_score(avg_accuracy, int(avg_moral_points))
        
        # Create synthetic team user
        from aimodelshare.moral_compass.api_client import MoralcompassApiClient
        
        api_client = MoralcompassApiClient()
        team_username = f"team:{team_name}"
        
        # Update team entry
        response = api_client.update_moral_compass(
            table_id=table_id,
            username=team_username,
            metrics={
                'accuracy': avg_accuracy,
                'moral_points': avg_moral_points,
                'combined_score': combined_score,
                'member_count': len(team_data['members'])
            },
            tasks_completed=0,
            total_tasks=0,
            questions_correct=0,
            total_questions=0,
            primary_metric='combined_score'
        )
        
        server_score = response.get('moralCompassScore', combined_score)
        
        logger.info(
            f"Team sync successful: team={team_name}, members={len(team_data['members'])}, "
            f"avg_accuracy={avg_accuracy:.4f}, avg_moral={avg_moral_points:.1f}, "
            f"combined={combined_score:.4f}, server_score={server_score:.4f}"
        )
        
        return {
            'synced': True,
            'status': 'synced',
            'server_score': server_score,
            'local_preview': combined_score,
            'message': f'✓ Team synced! Score: {server_score:.4f}'
        }
        
    except Exception as e:
        logger.error(f"Team sync failed for '{team_name}': {e}")
        return {
            'synced': False,
            'status': 'error',
            'error': str(e),
            'message': f'⚠️ Team sync error: {str(e)}'
        }


def _aggregate_team_data(team_name: str, table_id: str) -> Dict[str, Any]:
    """
    Aggregate data for all team members.
    
    Args:
        team_name: The team name
        table_id: The table ID
        
    Returns:
        Dictionary with:
        - 'members': List of member usernames
        - 'avg_accuracy': Average accuracy across members
        - 'avg_moral_points': Average moral points across members
    """
    try:
        # Get team members from environment or use heuristic
        team_members = _get_team_members(team_name)
        
        if not team_members:
            logger.warning(f"No team members configured for team '{team_name}'")
            return {'members': [], 'avg_accuracy': 0.0, 'avg_moral_points': 0.0}
        
        # Fetch accuracy data from playground
        accuracy_data = _fetch_team_accuracy_data(team_members)
        
        # Fetch moral compass data
        moral_data = _fetch_team_moral_data(team_members, table_id)
        
        # Compute averages
        valid_members = set(accuracy_data.keys()) & set(moral_data.keys())
        
        if not valid_members:
            return {'members': [], 'avg_accuracy': 0.0, 'avg_moral_points': 0.0}
        
        avg_accuracy = sum(accuracy_data[m] for m in valid_members) / len(valid_members)
        avg_moral = sum(moral_data[m] for m in valid_members) / len(valid_members)
        
        return {
            'members': list(valid_members),
            'avg_accuracy': avg_accuracy,
            'avg_moral_points': avg_moral
        }
        
    except Exception as e:
        logger.error(f"Failed to aggregate team data: {e}")
        return {'members': [], 'avg_accuracy': 0.0, 'avg_moral_points': 0.0}


def _get_team_members(team_name: str) -> List[str]:
    """
    Get list of team members.
    
    Args:
        team_name: The team name
        
    Returns:
        List of member usernames
        
    Note:
        Currently reads from TEAM_MEMBERS environment variable (comma-separated).
        Future enhancement: read from team registry or user profiles.
    """
    # Check environment variable
    members_str = os.getenv('TEAM_MEMBERS', '')
    if members_str:
        return [m.strip() for m in members_str.split(',') if m.strip()]
    
    # Fallback: try to infer from current user
    username = os.getenv('username')
    if username:
        return [username]
    
    return []


def _fetch_team_accuracy_data(members: List[str]) -> Dict[str, float]:
    """
    Fetch accuracy data for team members from playground.
    
    Args:
        members: List of member usernames
        
    Returns:
        Dictionary mapping username -> accuracy
    """
    try:
        from aimodelshare.playground import Competition
        
        playground_url = os.getenv('PLAYGROUND_URL',
                                   'https://cf3wdpkg0d.execute-api.us-east-1.amazonaws.com/prod/m')
        
        playground = Competition(playground_url)
        leaderboard = playground.get_leaderboard()
        
        accuracy_data = {}
        for entry in leaderboard:
            username = entry.get('username') or entry.get('user')
            if username in members:
                accuracy = (
                    entry.get('accuracy') or
                    entry.get('test_accuracy') or
                    entry.get('score', 0.7)
                )
                accuracy_data[username] = float(accuracy)
        
        return accuracy_data
        
    except Exception as e:
        logger.error(f"Failed to fetch team accuracy data: {e}")
        return {}


def _fetch_team_moral_data(members: List[str], table_id: str) -> Dict[str, float]:
    """
    Fetch moral compass data for team members.
    
    Args:
        members: List of member usernames
        table_id: The table ID
        
    Returns:
        Dictionary mapping username -> moral_points
    """
    try:
        from aimodelshare.moral_compass.api_client import MoralcompassApiClient
        
        api_client = MoralcompassApiClient()
        
        moral_data = {}
        for username in members:
            try:
                user_stats = api_client.get_user(table_id, username)
                # Extract moral points from moralCompassScore (reverse normalization estimate)
                # This is an approximation; ideally we'd store raw points separately
                moral_score = user_stats.total_count if hasattr(user_stats, 'total_count') else 0
                moral_data[username] = float(moral_score)
            except Exception as e:
                logger.debug(f"Could not fetch moral data for {username}: {e}")
                continue
        
        return moral_data
        
    except Exception as e:
        logger.error(f"Failed to fetch team moral data: {e}")
        return {}


# ============================================================================
# Leaderboard Cache & Generation
# ============================================================================

# Global cache for leaderboard data
_leaderboard_cache: Dict[str, Tuple[float, List[Dict[str, Any]]]] = {}


def fetch_cached_users(table_id: Optional[str] = None, ttl: int = 30) -> List[Dict[str, Any]]:
    """
    Fetch users from moral compass table with caching.
    
    Args:
        table_id: Optional table ID (auto-derived if not provided)
        ttl: Cache TTL in seconds (default: 30)
        
    Returns:
        List of user dictionaries with fields:
        - 'username': str
        - 'moralCompassScore': float
        - 'submissionCount': int (if available)
        - 'totalCount': int (if available)
    """
    if not table_id:
        table_id = _derive_table_id()
    
    # Check cache
    cache_key = table_id
    if cache_key in _leaderboard_cache:
        cache_time, cached_data = _leaderboard_cache[cache_key]
        if (time.time() - cache_time) < ttl:
            logger.debug(f"Using cached leaderboard for table {table_id}")
            return cached_data
    
    # Fetch from API
    try:
        from aimodelshare.moral_compass.api_client import MoralcompassApiClient
        
        api_client = MoralcompassApiClient()
        users = list(api_client.iter_users(table_id))
        
        # Convert to dict format
        user_list = []
        for user in users:
            user_list.append({
                'username': user.username,
                'moralCompassScore': user.total_count,  # Assuming total_count stores combined score
                'submissionCount': user.submission_count,
                'totalCount': user.total_count
            })
        
        # Update cache
        _leaderboard_cache[cache_key] = (time.time(), user_list)
        
        logger.info(f"Fetched {len(user_list)} users for table {table_id}")
        return user_list
        
    except Exception as e:
        logger.error(f"Failed to fetch users for leaderboard: {e}")
        return []


def build_moral_leaderboard_html(
    highlight_username: Optional[str] = None,
    include_teams: bool = True,
    table_id: Optional[str] = None,
    max_entries: int = 20
) -> str:
    """
    Build HTML for moral compass leaderboard.
    
    Args:
        highlight_username: Username to highlight (current user)
        include_teams: If True, include team entries
        table_id: Optional table ID (auto-derived if not provided)
        max_entries: Maximum number of entries to display
        
    Returns:
        HTML string with leaderboard table
        
    Note:
        Uses same styling classes as model_building_game:
        - leaderboard-html-table
        - user-row-highlight
    """
    users = fetch_cached_users(table_id)
    
    if not users:
        return """
        <div style='text-align: center; padding: 40px; color: var(--text-muted);'>
            <p>No leaderboard data available yet.</p>
            <p>Complete activities and sync to appear on the leaderboard!</p>
        </div>
        """
    
    # Filter teams if needed
    if not include_teams:
        users = [u for u in users if not u['username'].startswith('team:')]
    
    # Sort by moralCompassScore descending
    users_sorted = sorted(users, key=lambda u: u['moralCompassScore'], reverse=True)
    users_sorted = users_sorted[:max_entries]
    
    # Build HTML
    html = """
    <table class='leaderboard-html-table'>
        <thead>
            <tr>
                <th>Rank</th>
                <th>Name</th>
                <th>Moral Compass Score</th>
                <th>Type</th>
            </tr>
        </thead>
        <tbody>
    """
    
    for rank, user in enumerate(users_sorted, start=1):
        username = user['username']
        score = user['moralCompassScore']
        
        is_team = username.startswith('team:')
        display_name = username[5:] if is_team else username  # Remove 'team:' prefix
        entry_type = '👥 Team' if is_team else '👤 User'
        
        # Highlight current user
        highlight = username == highlight_username
        row_class = "class='user-row-highlight'" if highlight else ""
        
        html += f"""
            <tr {row_class}>
                <td>{rank}</td>
                <td>{display_name}</td>
                <td>{score:.4f}</td>
                <td>{entry_type}</td>
            </tr>
        """
    
    html += """
        </tbody>
    </table>
    """
    
    return html


# ============================================================================
# Convenience Functions
# ============================================================================

def get_moral_compass_widget_html(
    local_points: int,
    server_score: Optional[float] = None,
    is_synced: bool = False
) -> str:
    """
    Generate HTML for Moral Compass widget.
    
    Args:
        local_points: Local moral points accumulated
        server_score: Server moral compass score (if synced)
        is_synced: Whether currently synced
        
    Returns:
        HTML string for widget display
    """
    status_icon = "✓" if is_synced else "⏳"
    status_text = "(synced)" if is_synced else "(pending)"
    
    html = f"""
    <div style='background: var(--block-background-fill); padding: 16px; border-radius: 8px; 
                border: 2px solid var(--accent-strong); margin: 16px 0;'>
        <h3 style='margin-top: 0;'>🧭 Moral Compass Score</h3>
        <div style='display: flex; justify-content: space-around; flex-wrap: wrap;'>
            <div style='text-align: center; margin: 10px;'>
                <div style='font-size: 0.9rem; color: var(--text-muted);'>Local Points</div>
                <div style='font-size: 2rem; font-weight: bold; color: var(--accent-strong);'>
                    {local_points}
                </div>
            </div>
    """
    
    if server_score is not None:
        html += f"""
            <div style='text-align: center; margin: 10px;'>
                <div style='font-size: 0.9rem; color: var(--text-muted);'>Server Score {status_icon}</div>
                <div style='font-size: 2rem; font-weight: bold; color: var(--accent-strong);'>
                    {server_score:.4f}
                </div>
                <div style='font-size: 0.8rem; color: var(--text-muted);'>{status_text}</div>
            </div>
        """
    
    html += """
        </div>
    </div>
    """
    
    return html



"""Optional dependency checking utilities."""
import os
import importlib.util
import warnings

_DEF_SUPPRESS_ENV = "AIMODELSHARE_SUPPRESS_OPTIONAL_WARNINGS"


def check_optional(name: str, feature_label: str, suppress_env: str = _DEF_SUPPRESS_ENV) -> bool:
    """Check if an optional dependency is available.
    
    Print a single warning (via warnings) if missing and suppression env var is not set.
    Returns True if available, False otherwise.
    
    Parameters
    ----------
    name : str
        The name of the module to check (e.g., 'xgboost', 'pyspark')
    feature_label : str
        A human-readable label for the feature that requires this dependency
    suppress_env : str, optional
        Environment variable name to check for suppression (default: AIMODELSHARE_SUPPRESS_OPTIONAL_WARNINGS)
    
    Returns
    -------
    bool
        True if the module is available, False otherwise
    """
    spec = importlib.util.find_spec(name)
    if spec is None:
        if not os.environ.get(suppress_env):
            warnings.warn(
                f"{feature_label} support unavailable. Install `{name}` to enable.",
                category=UserWarning,
                stacklevel=2,
            )
        return False
    return True



import email
import boto3
import botocore
import os
import jwt
from numpy.core.fromnumeric import var
import requests
import uuid
import json
import math
import time
import datetime
import onnx
import tempfile
import types
import sys
import base64
import mimetypes
import numpy as np
import pandas as pd
from aimodelshare.tools import extract_varnames_fromtrainingdata, _get_extension_from_filepath
from aimodelshare.aws import get_s3_iam_client, run_function_on_lambda, get_token, get_aws_token, get_aws_client
from aimodelshare.bucketpolicy import _custom_upload_policy
from aimodelshare.exceptions import AuthorizationError, AWSAccessError, AWSUploadError
from aimodelshare.api import get_api_json
from aimodelshare.modeluser import decode_token_unverified
from aimodelshare.preprocessormodules import upload_preprocessor
from aimodelshare.model import _get_predictionmodel_key, _extract_model_metadata
from aimodelshare.data_sharing.share_data import share_data_codebuild
from aimodelshare.aimsonnx import _get_metadata
from aimodelshare.utils import HiddenPrints

def take_user_info_and_generate_api(model_filepath, model_type, categorical,labels, preprocessor_filepath,
                                    custom_libraries, requirements, exampledata_json_filepath, repo_name, 
                                    image_tag, reproducibility_env_filepath, memory, timeout, pyspark_support=False):
    """
    Generates an api using model parameters and user credentials, from the user

    -----------
    Parameters

    model_filepath :  string ends with '.onnx'
                    value - Absolute path to model file [REQUIRED] to be set by the user
                    .onnx is the only accepted model file extension
                    "example_model.onnx" filename for file in directory.
                    "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
    preprocessor_filepath:  string
                            value - absolute path to preprocessor file 
                            [REQUIRED] to be set by the user
                            "./preprocessor.zip" 
                            searches for an exported zip preprocessor file in the current directory
                            file is generated using export_preprocessor function from the AI Modelshare library 
    model_type :  string 
                values - [ 'text' , 'image' , 'tabular' , 'timeseries' ]
                Type of model data 
    categorical:  string
                  "TRUE" if model is Classification, categorical type
                  "FALSE" if model is continuous, Regression type
    labels:   list
            value - labels for training data
            can be extracted from columns of y train or can be provided by the user
    custom_libraries:   string
                  "TRUE" if user wants to load custom Python libraries to their prediction runtime
                  "FALSE" if user wishes to use AI Model Share base libraries including latest versions of most common ML libs.
     

    -----------
    Returns
    finalresult : list 
                length 5
                [ api url, api status code,   
                start time for api generation, 
                model unique key, 
                s3 bucket name
                  ]                   

    """
    import tempfile
    from zipfile import ZipFile
    import os
    import random

    # create temporary folder
    temp_dir = tempfile.gettempdir()
   
    api_json= get_api_json()
    user_client = boto3.client('apigateway', aws_access_key_id=str(
    os.environ.get("AWS_ACCESS_KEY_ID_AIMS")), aws_secret_access_key=str(os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS")), region_name=str(os.environ.get("AWS_REGION_AIMS")))

    response2 = user_client.import_rest_api(
    failOnWarnings=True,
    parameters={
        'endpointConfigurationTypes': 'REGIONAL'
    },
    body=api_json
    )

    api_id = response2['id']
    now = datetime.datetime.now()
    s3, iam, region = get_s3_iam_client(os.environ.get("AWS_ACCESS_KEY_ID_AIMS"), os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), os.environ.get("AWS_REGION_AIMS"))
    
    def create_bucket(s3_client, bucket_name, region):
        try:
            response=s3_client.head_bucket(Bucket=bucket_name)
        except:
            if(region=="us-east-1"):
                response = s3_client.create_bucket(
                    ACL="private",
                    Bucket=bucket_name
                )
            else:
                location={'LocationConstraint': region}
                response=s3_client.create_bucket(
                    ACL="private",
                    Bucket=bucket_name,
                    CreateBucketConfiguration=location
                )
        return response

    create_bucket(s3['client'], os.environ.get("BUCKET_NAME"), region)

    # model upload
    if isinstance(model_filepath, onnx.ModelProto):
        model = model_filepath
        temp_prep=tempfile.mkdtemp()
        Filepath = temp_prep+"/model.onnx"
        with open(Filepath, "wb") as f:
            f.write(model.SerializeToString())
    else:
        Filepath = model_filepath
        model = onnx.load(model_filepath)
    metadata = _extract_model_metadata(model)
    input_shape = metadata["input_shape"]
    #tab_imports ='./tabular_imports.pkl'
    #img_imports ='./image_imports.pkl'
    file_extension = _get_extension_from_filepath(Filepath)
    unique_model_id = str(api_id)
    file_key, versionfile_key = _get_predictionmodel_key(
        unique_model_id, file_extension)
    try:
        s3["client"].upload_file(exampledata_json_filepath, os.environ.get("BUCKET_NAME"), unique_model_id + "/exampledata.json") 
    except:
        pass
    try:
        s3["client"].upload_file(Filepath, os.environ.get("BUCKET_NAME"),  file_key)
        s3["client"].upload_file(Filepath, os.environ.get("BUCKET_NAME"),  versionfile_key)

        # preprocessor upload
        #s3["client"].upload_file(tab_imports, os.environ.get("BUCKET_NAME"),  'tabular_imports.pkl')
        #s3["client"].upload_file(img_imports, os.environ.get("BUCKET_NAME"),  'image_imports.pkl')
        # preprocessor upload

        # ADD model/Preprocessor VERSION
        if isinstance(preprocessor_filepath, types.FunctionType): 
            from aimodelshare.preprocessormodules import export_preprocessor
            temp_prep=tempfile.mkdtemp()
            with HiddenPrints():
                export_preprocessor(preprocessor_filepath,temp_prep)
            preprocessor_filepath = temp_prep+"/preprocessor.zip"
        response = upload_preprocessor(
            preprocessor_filepath, s3, os.environ.get("BUCKET_NAME"), unique_model_id, 1)
        preprocessor_file_extension = _get_extension_from_filepath(
            preprocessor_filepath)
        # write runtime JSON
        json_path = os.path.join(temp_dir, "runtime_data.json")
        if(preprocessor_file_extension == '.py'):
            runtime_preprocessor_type = "module"
        elif(preprocessor_file_extension == '.pkl'):
            runtime_preprocessor_type = "pickle object"
        else:
            runtime_preprocessor_type = "others"
        runtime_data = {}
        runtime_data["runtime_model"] = {"name": "runtime_model.onnx"}
        runtime_data["runtime_preprocessor"] = runtime_preprocessor_type
            
        #runtime_data = {"runtime_model": {"name": "runtime_model.onnx"},"runtime_preprocessor": runtime_preprocessor_type }
        json_string = json.dumps(runtime_data, sort_keys=False)
        with open(json_path, 'w') as outfile:
            outfile.write(json_string)
        s3["client"].upload_file(
            json_path, os.environ.get("BUCKET_NAME"), unique_model_id + "/runtime_data.json"
        )
        os.remove(json_path)

        # upload model metadata
        upload_model_metadata(model, s3, os.environ.get("BUCKET_NAME"), unique_model_id)

        # upload reproducibility env
        if reproducibility_env_filepath:
            upload_reproducibility_env(reproducibility_env_filepath, s3, os.environ.get("BUCKET_NAME"), unique_model_id)
    except Exception as err:
        raise AWSUploadError(
            "There was a problem with model/preprocessor upload. "+str(err))

    #Delete Legacy exampledata json:
    try:
        os.remove(exampledata_json_filepath)
    except:
        pass

    #headers = {'content-type': 'application/json'}

    ### Progress Update #2/6 {{{
    sys.stdout.write('\r')
    sys.stdout.write("[========                             ] Progress: 30% - Building lambda functions and updating permissions...")
    sys.stdout.flush()
    # }}}
    
    from aimodelshare.api import create_prediction_api
    apiurl = create_prediction_api(model_filepath, unique_model_id,
                                   model_type, categorical, labels,api_id,
                                   custom_libraries, requirements, repo_name, 
                                   image_tag, memory, timeout, pyspark_support=pyspark_support)

    finalresult = [apiurl["body"], apiurl["statusCode"],
                   now, unique_model_id, os.environ.get("BUCKET_NAME"), input_shape]
    return finalresult

def upload_reproducibility_env(reproducibility_env_file, s3, bucket, model_id):
    # Check the reproducibility_env {{{
    with open(reproducibility_env_file) as json_file:
      reproducibility_env = json.load(json_file)
      if "global_seed_code" not in reproducibility_env \
              or "local_seed_code" not in reproducibility_env \
              or "gpu_cpu_parallelism_ops" not in reproducibility_env \
              or "session_runtime_info" not in reproducibility_env:
          raise Exception("reproducibility environment is not complete")

    # Upload the json {{{
    try:
        s3["client"].upload_file(
            reproducibility_env_file, bucket, model_id + "/runtime_reproducibility.json"
        )
    except Exception as err:
        raise err
    # }}}

def upload_model_metadata(model, s3, bucket, model_id):
    meta_dict = _get_metadata(model)
    model_metadata = {
        "model_config": meta_dict["model_config"],
        "ml_framework": meta_dict["ml_framework"],
        "model_type": meta_dict["model_type"]
    }

    temp = tempfile.mkdtemp()
    model_metadata_path = temp + "/" + 'model_metadata.json'
    with open(model_metadata_path, 'w') as outfile:
        json.dump(model_metadata, outfile)

    # Upload the json {{{
    try:
        s3["client"].upload_file(
            model_metadata_path, bucket, model_id + "/runtime_metadata.json"
        )
    except Exception as err:
        raise err

def send_model_data_to_dyndb_and_return_api(api_info, private, categorical, preprocessor_filepath,
                                            aishare_modelname, aishare_modeldescription, aishare_modelevaluation, model_type,
                                            aishare_tags, aishare_apicalls, exampledata_json_filepath,
                                            variablename_and_type_data="default", email_list=[]):
    """
    Updates dynamodb with model data taken as input from user along with already generated api info
    -----------
    Parameters
    api_info :  list  
              length 5
              api and s3 bucket information 
              returned from take_user_info_and_generate_api function
    private :   string, default="FALSE"
              TRUE if model and its corresponding data is not public
              FALSE if model and its corresponding data is public            
    categorical: string, default="TRUE"
                  TRUE if model is of Classification type with categorical variables
                  FALSE if model is Regression type with continuous variables
    preprocessor_filepath:  string
                          value - Absolute path to preprocessor file [REQUIRED] to be set by the user
                          "./preprocessor.zip" 
                          searches for an exported zip preprocessor file in the current directory 
    variablename_and_type_data :  list, default='default'
                                value- extracted from example_data
                                [variable types,variable columns]
                                'default' when training data info is not available to extract columns
    email_list: list of strings
                values - list including all emails of users who have access the playground.
                list should contain same emails used to sign up for modelshare.ai account.
                [OPTIONAL] to be set by the playground owner
    -----------
    Results
    print (api_info) : statements with the generated live prediction API information for the user

    """
    
    # unpack user credentials
    unique_model_id = api_info[3]
    bucket_name = api_info[4]
    input_shape = api_info[5]
    if variablename_and_type_data == "default":
        variablename_and_type_data = ["", ""]

     # needs to use double backslashes and have full filepath

    if isinstance(preprocessor_filepath, types.FunctionType): 
        from aimodelshare.preprocessormodules import export_preprocessor
        temp_prep=tempfile.mkdtemp()
        with HiddenPrints():
            export_preprocessor(preprocessor_filepath,temp_prep)
        preprocessor_filepath = temp_prep+"/preprocessor.zip"
    preprocessor_file_extension = _get_extension_from_filepath(
        preprocessor_filepath)
    if exampledata_json_filepath!="":
        exampledata_addtodatabase={"exampledata":"TRUE"}
    else:
        exampledata_addtodatabase={"exampledata":"FALSE"}
    bodydata = {
        "id": int(math.log(1/((time.time()*1000000)))*100000000000000),
        "unique_model_id": unique_model_id,
        "apideveloper": os.environ.get("username"),  # change this to first and last name
        "apimodeldescription": aishare_modeldescription,
        "apimodelevaluation": aishare_modelevaluation,
        "apimodeltype": model_type,
        # getting rid of extra quotes that screw up dynamodb string search on apiurls
        "apiurl": api_info[0].strip('\"'),
        "bucket_name": bucket_name,
        "version": 1,
        "modelname": aishare_modelname,
        "tags": aishare_tags,
        "Private": private,
        "Categorical": categorical,
        "delete": "FALSE",
        "input_feature_dtypes": variablename_and_type_data[0],
        "input_feature_names": variablename_and_type_data[1],
        "preprocessor": preprocessor_filepath,
        "preprocessor_fileextension": preprocessor_file_extension,
        "input_shape": input_shape,
        "email_list": email_list,
        "useremails": ','.join(email_list),
    }
    bodydata.update(exampledata_addtodatabase)
    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    response = requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                              json=bodydata, headers=headers_with_authentication)
    response_string = response.text
    response_string = response_string[1:-1]
    import json
    response_stringfinal = json.loads(response_string).get("id","no id")
    # Build output {{{
    final_message = ("\nYou can now use your Model Playground.\n\n"
                     "Follow this link to explore your Model Playground's functionality\n"
                     "You can make predictions with the Dashboard and access example code from the Programmatic tab.\n")
    web_dashboard_url = ("https://www.modelshare.ai/detail/"+ response_stringfinal)
    
    start = api_info[2]
    end = datetime.datetime.now()
    difference = (end - start).total_seconds()
    finalresult2 = "Success! Your Model Playground was created in " + \
        str(int(difference)) + " seconds. \n" + " Playground Url: " + api_info[0] 


    # }}}

    ### Progress Update #6/6 {{{
    sys.stdout.write('\r')
    sys.stdout.write("[=====================================] Progress: 100% - Complete!                                            ")
    sys.stdout.flush()
    # }}}

    return print("\n\n" + finalresult2 + "\n" + final_message + web_dashboard_url)


def model_to_api(model_filepath, model_type, private, categorical, y_train, preprocessor_filepath, 
                custom_libraries="FALSE", example_data=None, image="", 
                base_image_api_endpoint="https://vupwujn586.execute-api.us-east-1.amazonaws.com/dev/copybasetouseracct", 
                update=False, reproducibility_env_filepath=None, memory=None, timeout=None, email_list=[],pyspark_support=False,
                input_dict=None, print_output=True, playground_id=False):
    """
      Launches a live prediction REST API for deploying ML models using model parameters and user credentials, provided by the user
      Inputs : 8
      Output : model launched to an API
               detaled API info printed out
     
      -----------
      Parameters 
      
      model_filepath :  string ends with '.onnx'
                        value - Absolute path to model file 
                        [REQUIRED] to be set by the user
                        .onnx is the only accepted model file extension
                        "example_model.onnx" filename for file in directory.
                        "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
      preprocessor_filepath:  string
                              value - absolute path to preprocessor file 
                              [REQUIRED] to be set by the user
                              "./preprocessor.zip" 
                              searches for an exported zip preprocessor file in the current directory
                              file is generated using export_preprocessor function from the AI Modelshare library  
      model_type :  string 
                    values - [ 'Text' , 'Image' , 'Tabular' , 'Timeseries' ] 
                    type of model data     
      categorical:    bool, default=True
                      True [DEFAULT] if model is of Classification type with categorical target variables
                      False if model is of Regression type with continuous target variables
      y_train : training labels of size of dataset
                value - y values for model
                [REQUIRED] for classification type models
                expects a one hot encoded y train data format
      private :   bool, default = False
                  True if model and its corresponding data is not public
                  False [DEFAULT] if model and its corresponding data is public   
      custom_libraries:   string
                  "TRUE" if user wants to load custom Python libraries to their prediction runtime
                  "FALSE" if user wishes to use AI Model Share base libraries including latest versions of most common ML libs.
      example_data:  pandas DataFrame (for tabular & text data) OR filepath as string (image, audio, video data)
                     tabular data - pandas DataFrame in same structure expected by preprocessor function
                     other data types - absolute path to folder containing example data
                                        (first five files with relevent file extensions will be accepted)
                     [REQUIRED] for tabular data
      reproducibility_env_filepath: string
                                value - absolute path to environment environment json file 
                                [OPTIONAL] to be set by the user
                                "./reproducibility.json" 
                                file is generated using export_reproducibility_env function from the AI Modelshare library
      email_list: list of strings
                values - list including all emails of users who have access the playground.
                list should contain same emails used to sign up for modelshare.ai account.
                [OPTIONAL] to be set by the playground owner
      -----------
      Returns
      print_api_info : prints statements with generated live prediction API details
                      also prints steps to update the model submissions by the user/team
                                 

    """
    # Used 2 python functions :
    #  1. take_user_info_and_generate_api : to upload model/preprocessor and generate an api for model submitted by user
    #  2. send_model_data_to_dyndb_and_return_api : to add new record to database with user data, model and api related information


    # Get user inputs, pass to other functions  {{{
    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID_AIMS"),
                                         aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), 
                                         region_name=os.environ.get("AWS_REGION_AIMS"))

    if all([isinstance(email_list, list)]):
        idtoken = get_aws_token()
        decoded = decode_token_unverified(idtoken)

        email=None
        email = decoded['email']
        # Owner has to be the first on the list
        email_list.insert(0, email)
        if any([private==False,private==None]):
          email_list=["publicaimsplayground"]
        else:
          pass
    else:
        return print("email_list argument empty or incorrectly formatted. Please provide a list of emails for authorized playground users formatted as strings.")

    if(image!=""):
        repo_name, image_tag = image.split(':')
    elif model_type=="tabular":
        repo_name, image_tag = "aimodelshare_base_image", "tabular"
    elif model_type=="text":
        repo_name, image_tag = "aimodelshare_base_image", "texttest"
    elif model_type=="image":
        repo_name, image_tag = "aimodelshare_base_image", "v3"
    elif model_type=="video":
        repo_name, image_tag = "aimodelshare_base_image", "v3"
    else:
        repo_name, image_tag = "aimodelshare_base_image", "v3"

    # Pyspark mode
    if pyspark_support:
        repo_name, image_tag = "aimodelshare_base_image", "pyspark"
    
    from aimodelshare.containerization import clone_base_image
    response = clone_base_image(user_session, repo_name, image_tag, "517169013426", base_image_api_endpoint, update)
    if(response["Status"]==0):
        print(response["Success"])
        return

    if input_dict == None: 
        print("We need some information about your model before we can build your REST API and interactive Model Playground.")
        print("   ")

        requirements = ""
        if(any([custom_libraries=='TRUE',custom_libraries=='true'])):
            requirements = input("Enter all required Python libraries you need at prediction runtime (separated with commas):")
            #_confirm_libraries_exist(requirements)
            
        aishare_modelname = input("Model Name (for AI Model Share Website):")
        aishare_modeldescription = input("Model Description (Explain what your model does and \n why end-users would find your model useful):")

        aishare_tags = input(
            "Model Key Words (Search categories that describe your model, separated with commas):")
        print("   ")
        #  }}}
    else: 
        requirements = input_dict.get("requirements","")
        aishare_modelname = input_dict.get("model_name","")
        aishare_modeldescription = input_dict.get("model_description","")
        aishare_tags = input_dict.get("tags","")

    aishare_modelevaluation = "unverified" # verified metrics added to playground once 1. a model is submitted to a competition leaderboard and 2. playground owner updates runtime
                                           #...model with update_runtime_model()
    aishare_apicalls = 0

    # Force user to provide example data for tabular models {{{
    if any([model_type.lower() == "tabular", model_type.lower() == "timeseries"]):
        if not isinstance(example_data, pd.DataFrame):
            return print("Error: Example data is required for tabular models. \n Please provide a pandas DataFrame with a sample of your X data (in the format expected by your preprocessor) and try again.")
    else:
        pass
    #}}}
        
    print("Creating your prediction API. (This process may take several minutes.)\n")
    variablename_and_type_data = None
    private = str(private).upper()
    categorical = str(categorical).upper()
    if model_type == "tabular" or "keras_tabular" or 'Tabular':
        variablename_and_type_data = extract_varnames_fromtrainingdata(
            example_data)
    if categorical == "TRUE":
        try:
            labels = y_train.columns.tolist()
        except:
            #labels = list(set(y_train.to_frame()['tags'].tolist()))
            labels = list(set(y_train))
    else:
        labels = "no data"
    temp_dir = tempfile.gettempdir()
    # Create Example Data JSON
    exampledata_json_filepath = ""
    if example_data is not None:
        _create_exampledata_json(model_type, example_data)
        exampledata_json_filepath = temp_dir+ "/exampledata.json"

    ### Progress Update #1/6 {{{
    sys.stdout.write("[===                                  ] Progress: 5% - Accessing Amazon Web Services, uploading resources...")
    sys.stdout.flush()
    # }}}

    api_info = take_user_info_and_generate_api( 
        model_filepath, model_type, categorical, labels, 
        preprocessor_filepath, custom_libraries, requirements, 
        exampledata_json_filepath, repo_name, image_tag, 
        reproducibility_env_filepath, memory, timeout, pyspark_support=pyspark_support)

    if input_dict and playground_id:
        aishare_modelname = aishare_modelname+" "+str(api_info[3])

    ### Progress Update #5/6 {{{
    sys.stdout.write('\r')
    sys.stdout.write("[=================================    ] Progress: 90% - Finishing web dashboard...                           ")
    sys.stdout.flush()
    # }}}
    
    print_api_info = send_model_data_to_dyndb_and_return_api(
        api_info, private, categorical,preprocessor_filepath, aishare_modelname,
        aishare_modeldescription, aishare_modelevaluation, model_type,
        aishare_tags, aishare_apicalls, exampledata_json_filepath,
        variablename_and_type_data, email_list)
    
    return api_info[0]

def create_competition(apiurl, data_directory, y_test, eval_metric_filepath=None, email_list=[], public=False, public_private_split=0.5,
                        input_dict=None, print_output=True):
    """
    Creates a model competition for a deployed prediction REST API
    Inputs : 4
    Output : Create ML model competition and allow authorized users to submit models to resulting leaderboard/competition
    
    ---------
    Parameters
    
    apiurl: string
            URL of deployed prediction API 
    
    y_test :  list of y values for test data used to generate metrics from predicted values from X test data submitted via the submit_model() function
                [REQUIRED] to generate eval metrics in competition leaderboard
            
    data_directory : folder storing training data and test data (excluding Y test data)
    email_list: [REQUIRED] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
                                          
    ---------
    Returns
    finalmessage : Information such as how to submit models to competition

    """
    if all([isinstance(email_list, list)]):
        if any([len(email_list)>0, public=="True",public=="TRUE",public==True]):
              idtoken=get_aws_token()
              decoded = decode_token_unverified(idtoken)

              email=decoded['email']
              email_list.append(email)
    else:
        return print("email_list argument empty or incorrectly formatted.  Please provide a list of emails for authorized competition participants formatted as strings.")

    # create temporary folder
    temp_dir = tempfile.gettempdir()
    
    s3, iam, region = get_s3_iam_client(os.environ.get("AWS_ACCESS_KEY_ID_AIMS"), os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), os.environ.get("AWS_REGION_AIMS"))
    
    # Get bucket and model_id subfolder for user based on apiurl {{{
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
    # }}} 
    
    # upload y_test data: 
    ytest_path = os.path.join(temp_dir, "ytest.pkl")
    import pickle
    #ytest data to load to s3

    if y_test is not None:
        if type(y_test) is not list:
            y_test=y_test.tolist()
        else: 
            pass

        if all(isinstance(x, (np.float64)) for x in y_test):
              y_test = [float(i) for i in y_test]
        else: 
            pass


    pickle.dump(y_test,open(ytest_path,"wb"))
    s3["client"].upload_file(ytest_path, os.environ.get("BUCKET_NAME"),  model_id + "/competition/ytest.pkl")


    if eval_metric_filepath is not None:
    
        if isinstance(eval_metric_filepath, list):

            for i in eval_metric_filepath: 

                eval_metric_name = i.split('/')[-1]

                s3["client"].upload_file(i, os.environ.get("BUCKET_NAME"),  model_id + '/competition/metrics_' + eval_metric_name)

        else:

            eval_metric_name = eval_metric_filepath.split('/')[-1]

            print(eval_metric_name)

            s3["client"].upload_file(eval_metric_filepath, os.environ.get("BUCKET_NAME"), model_id + '/competition/metrics_' + eval_metric_name)


    # get api_id from apiurl, generate txt file name
    api_url_trim = apiurl.split('https://')[1]
    api_id = api_url_trim.split(".")[0]

    if input_dict == None:
        print("\n--INPUT COMPETITION DETAILS--\n")
        aishare_competitionname = input("Enter competition name:")
        aishare_competitiondescription = input("Enter competition description:")

        print("\n--INPUT DATA DETAILS--\n")
        print("Note: (optional) Save an optional LICENSE.txt file in your competition data directory to make users aware of any restrictions on data sharing/usage.\n")

        aishare_datadescription = input(
            "Enter data description (i.e.- filenames denoting training and test data, file types, and any subfolders where files are stored):")
        
        aishare_datalicense = input(
            "Enter optional data license descriptive name (e.g.- 'MIT, Apache 2.0, CC0, Other, etc.'):")

    else:
        aishare_competitionname = input_dict["competition_name"]
        aishare_competitiondescription = input_dict["competition_description"]
        aishare_datadescription = input_dict["data_description"]
        aishare_datalicense = input_dict["data_license"]


    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID_AIMS"),
                                          aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), 
                                         region_name=os.environ.get("AWS_REGION_AIMS"))

    account_number = user_session.client(
        'sts').get_caller_identity().get('Account')

    datauri=share_data_codebuild(account_number,os.environ.get("AWS_REGION_AIMS"),data_directory)
    
    #create and upload json file with list of authorized users who can submit to this competition.
    _create_competitionuserauth_json(apiurl, email_list,public,datauri['ecr_uri'], submission_type="competition")
    _create_public_private_split_json(apiurl, public_private_split, "competition")

    bodydata = {"unique_model_id": model_id,
                "bucket_name": api_bucket,
                "apideveloper": os.environ.get("username"),  # change this to first and last name
                "competitionname":aishare_competitionname,                
                "competitiondescription": aishare_competitiondescription,
                # getting rid of extra quotes that screw up dynamodb string search on apiurls
                "apiurl": apiurl,
                "version": 0,
                "Private": "FALSE",
                "delete": "FALSE",
                'datadescription':aishare_datadescription,
                'dataecruri':datauri['ecr_uri'],
                 'datalicense': aishare_datalicense}
    
    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)

      
    final_message = ("\n Success! Model competition created. \n\n"
                "You may now update your prediction API runtime model and verify evaluation metrics with the update_runtime_model() function.\n\n"
                "To upload new models and/or preprocessors to this API, team members should use \n"
                "the following credentials:\n\napiurl='" + apiurl+"'"+"\nfrom aimodelshare.aws import set_credentials\nset_credentials(apiurl=apiurl)\n\n"
                "They can then submit models to your competition by using the following code: \n\ncompetition= ai.Competition(apiurl)\n"
                "download_data('"+datauri['ecr_uri']+"') \n"
                 "# Use this data to preprocess data and train model. Write and save preprocessor fxn, save model to onnx file, generate predicted y values\n using X test data, then submit a model below.\n\n"
                "competition.submit_model(model_filepath, preprocessor_filepath, prediction_submission_list)")
  
    if print_output:
        return print(final_message)
    else:
        return



def create_experiment(apiurl, data_directory, y_test, eval_metric_filepath=None, email_list=[], public=False, public_private_split=0,
                        input_dict=None, print_output=True):
    """
    Creates a model experiment for a deployed prediction REST API
    Inputs : 4
    Output : Create ML model experiment and allow authorized users to submit models to resulting leaderboard/competition
    
    ---------
    Parameters
    
    apiurl: string
            URL of deployed prediction API 
    
    y_test :  list of y values for test data used to generate metrics from predicted values from X test data submitted via the submit_model() function
                [REQUIRED] to generate eval metrics in competition leaderboard
            
    data_directory : folder storing training data and test data (excluding Y test data)
    email_list: [REQUIRED] list of comma separated emails for users who are allowed to submit models to experiment.  Emails should be strings in a list.
                                          
    ---------
    Returns
    finalmessage : Information such as how to submit models to competition

    """
    if all([isinstance(email_list, list)]):
        if any([len(email_list)>0, public=="True",public=="TRUE",public==True]):
              idtoken=get_aws_token()
              decoded = decode_token_unverified(idtoken)

              email=decoded['email']
              email_list.append(email)
    else:
        return print("email_list argument empty or incorrectly formatted.  Please provide a list of emails for authorized competition participants formatted as strings.")

    # create temporary folder
    temp_dir = tempfile.gettempdir()
    
    s3, iam, region = get_s3_iam_client(os.environ.get("AWS_ACCESS_KEY_ID_AIMS"), os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), os.environ.get("AWS_REGION_AIMS"))
    
    # Get bucket and model_id subfolder for user based on apiurl {{{
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
    # }}} 
    
    # upload y_test data: 
    ytest_path = os.path.join(temp_dir, "ytest.pkl")
    import pickle
    #ytest data to load to s3

    if y_test is not None:
        if type(y_test) is not list:
            y_test=y_test.tolist()
        else: 
            pass

        if all(isinstance(x, (np.float64)) for x in y_test):
              y_test = [float(i) for i in y_test]
        else: 
            pass


    pickle.dump(y_test,open(ytest_path,"wb"))
    s3["client"].upload_file(ytest_path, os.environ.get("BUCKET_NAME"),  model_id + "/experiment/ytest.pkl")


    if eval_metric_filepath is not None:
    
        if isinstance(eval_metric_filepath, list):

            for i in eval_metric_filepath: 

                eval_metric_name = i.split('/')[-1]

                s3["client"].upload_file(i, os.environ.get("BUCKET_NAME"),  model_id + '/experiment/metrics_' + eval_metric_name)

        else:

            eval_metric_name = eval_metric_filepath.split('/')[-1]

            print(eval_metric_name)

            s3["client"].upload_file(eval_metric_filepath, os.environ.get("BUCKET_NAME"), model_id + '/experiment/metrics_' + eval_metric_name)


    # get api_id from apiurl, generate txt file name
    api_url_trim = apiurl.split('https://')[1]
    api_id = api_url_trim.split(".")[0]

    if input_dict == None:
        print("\n--INPUT EXPERIMENT DETAILS--\n")
        aishare_competitionname = input("Enter experiment name:")
        aishare_competitiondescription = input("Enter experiment description:")

        print("\n--INPUT DATA DETAILS--\n")
        print("Note: (optional) Save an optional LICENSE.txt file in your experiment data directory to make users aware of any restrictions on data sharing/usage.\n")

        aishare_datadescription = input(
            "Enter data description (i.e.- filenames denoting training and test data, file types, and any subfolders where files are stored):")
        
        aishare_datalicense = input(
            "Enter optional data license descriptive name (e.g.- 'MIT, Apache 2.0, CC0, Other, etc.'):")

    else:
        aishare_competitionname = input_dict["experiment_name"]
        aishare_competitiondescription = input_dict["experiment_description"]
        aishare_datadescription = input_dict["data_description"]
        aishare_datalicense = input_dict["data_license"]


    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID_AIMS"),
                                          aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), 
                                         region_name=os.environ.get("AWS_REGION_AIMS"))

    account_number = user_session.client(
        'sts').get_caller_identity().get('Account')

    datauri=share_data_codebuild(account_number,os.environ.get("AWS_REGION_AIMS"),data_directory)
    
    #create and upload json file with list of authorized users who can submit to this competition.
    _create_competitionuserauth_json(apiurl, email_list,public,datauri['ecr_uri'], submission_type="experiment")
    _create_public_private_split_json(apiurl, public_private_split, "experiment")

    bodydata = {"unique_model_id": model_id,
                "bucket_name": api_bucket,
                "apideveloper": os.environ.get("username"),  # change this to first and last name
                "experiment":"TRUE",
                "competitionname":aishare_competitionname,                
                "competitiondescription": aishare_competitiondescription,
                # getting rid of extra quotes that screw up dynamodb string search on apiurls
                "apiurl": apiurl,
                "version": 0,
                "Private": "FALSE",
                "delete": "FALSE",
                'datadescription':aishare_datadescription,
                'dataecruri':datauri['ecr_uri'],
                 'datalicense': aishare_datalicense}
    
    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)

      
    final_message = ("\n Success! Model experiment created. \n\n"
                "You may now update your prediction API runtime model and verify evaluation metrics with the update_runtime_model() function.\n\n"
                "To upload new models and/or preprocessors to this API, team members should use \n"
                "the following credentials:\n\napiurl='" + apiurl+"'"+"\nfrom aimodelshare.aws import set_credentials\nset_credentials(apiurl=apiurl)\n\n"
                "They can then submit models to your experiment by using the following code: \n\nexperiment= ai.Experiment(apiurl)\n"
                "download_data('"+datauri['ecr_uri']+"') \n"
                 "# Use this data to preprocess data and train model. Write and save preprocessor fxn, save model to onnx file, generate predicted y values\n using X test data, then submit a model below.\n\n"
                "experiment.submit_model(model_filepath, preprocessor_filepath, prediction_submission_list)")
  
    if print_output:
        return print(final_message)
    else:
        return

def _create_public_private_split_json(apiurl, split=0.5, submission_type='competition'): 
      import json
      if all(["AWS_ACCESS_KEY_ID_AIMS" in os.environ, 
            "AWS_SECRET_ACCESS_KEY_AIMS" in os.environ,
            "AWS_REGION_AIMS" in os.environ,
           "username" in os.environ, 
           "password" in os.environ]):
        pass
      else:
          return print("'Set public-private split' unsuccessful. Please provide credentials with set_credentials().")

      # Create user session
      aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                aws_region=os.environ.get('AWS_REGION_AIMS'))
      
      user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                        aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                        region_name=os.environ.get('AWS_REGION_AIMS'))
      
      s3 = user_sess.resource('s3')

      # Get bucket and model_id for user based on apiurl {{{
      response, error = run_function_on_lambda(
          apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
      )
      if error is not None:
          raise error

      _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
      # }}}
      
      import json  
      import tempfile
      tempdir = tempfile.TemporaryDirectory()
      with open(tempdir.name+'/public_private_split.json', 'w', encoding='utf-8') as f:
          json.dump({"public_private_split": str(split)}, f, ensure_ascii=False, indent=4)

      aws_client['client'].upload_file(
            tempdir.name+"/public_private_split.json", api_bucket, model_id +"/"+submission_type+"/public_private_split.json"
        )
      
      return

def _create_competitionuserauth_json(apiurl, email_list=[],public=False, datauri=None, submission_type="competition"): 
      import json
      if all(["AWS_ACCESS_KEY_ID_AIMS" in os.environ, 
            "AWS_SECRET_ACCESS_KEY_AIMS" in os.environ,
            "AWS_REGION_AIMS" in os.environ,
           "username" in os.environ, 
           "password" in os.environ]):
        pass
      else:
          return print("'Update Runtime Model' unsuccessful. Please provide credentials with set_credentials().")

      # Create user session
      aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                aws_region=os.environ.get('AWS_REGION_AIMS'))
      
      user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                        aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                        region_name=os.environ.get('AWS_REGION_AIMS'))
      
      s3 = user_sess.resource('s3')

      # Get bucket and model_id for user based on apiurl {{{
      response, error = run_function_on_lambda(
          apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
      )
      if error is not None:
          raise error

      _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
      # }}}

      
      import json  
      import tempfile
      tempdir = tempfile.TemporaryDirectory()
      with open(tempdir.name+'/competitionuserdata.json', 'w', encoding='utf-8') as f:
          json.dump({"emaillist": email_list, "public":str(public).upper(),"datauri":str(datauri)}, f, ensure_ascii=False, indent=4)

      aws_client['client'].upload_file(
            tempdir.name+"/competitionuserdata.json", api_bucket, model_id +"/"+submission_type+"/competitionuserdata.json"
        )
      
      return

def update_playground_access_list(apiurl, email_list=[], update_type="Add"): 
    """
    Updates list of authenticated participants who can submit new models to a competition.
    ---------------
    Parameters:
    apiurl: string
            URL of deployed prediction API 
    
    email_list: [REQUIRED] list of comma separated emails for users who are allowed to access model playground.  Emails should be strings in a list.
    update_type:[REQUIRED] options, string: 'Add', 'Remove', 'Replace'. Add appends user emails to original list, Remove deletes users from list, 
                and 'Replace' overwrites the original list with the new list provided.    
    -----------------
    Returns
    response:   "Success" upon successful request
    """
    if update_type not in ['Add', 'Remove', 'Replace']:
        return "Error: update_type must be in the form of 'Add', 'Remove', or 'Replace'"
    
    bodydata = {
        "modifyaccess": "TRUE",
        "apideveloper": os.environ.get("username"),  # change this to first and last name
        "apiurl": apiurl,
        "operation": update_type,
        "email_list": email_list,
    }

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    response = requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                            json=bodydata, headers=headers_with_authentication)
    response_string = response.text
    response_string = response_string[1:-1]

    # Build output {{{
    finalresult = "Success! Your Model Playground email list has been updated. Playground Url: " + apiurl
    return finalresult

def get_playground_access_list(apiurl):
    bodydata = {
        "getaccess": "TRUE",
        "apideveloper": os.environ.get("username"),  # change this to first and last name
        "apiurl": apiurl,
    }

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    response = requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                            json=bodydata, headers=headers_with_authentication)
    response_string = response.text
    response_string = response_string[1:-1]

    return response_string.split(',')

def update_access_list(apiurl, email_list=[],update_type="Add", submission_type="competition"):
      """
      Updates list of authenticated participants who can submit new models to a competition.
      ---------------
      Parameters:
      apiurl: string
              URL of deployed prediction API 
      
      email_list: [REQUIRED] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
      update_type:[REQUIRED] options, string: 'Add', 'Remove', 'Replace_list','Get. Add appends user emails to original list, Remove deletes users from list, 
                  'Replace_list' overwrites the original list with the new list provided, and Get returns the current list.    
      -----------------
      Returns
      response:   "Success" upon successful request
      """
      import json
      import os
      if all(["AWS_ACCESS_KEY_ID_AIMS" in os.environ, 
            "AWS_SECRET_ACCESS_KEY_AIMS" in os.environ,
            "AWS_REGION_AIMS" in os.environ,
           "username" in os.environ, 
           "password" in os.environ]):
        pass
      else:
          return print("'Update unsuccessful. Please provide credentials with set_credentials().")

      if all([isinstance(email_list, list)]):
          if all([len(email_list)>0]):
              pass
      else:
          return print("email_list argument empty or incorrectly formatted.  Please provide a list of emails for authorized competition participants formatted as strings.")

      # Create user session
      aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                aws_region=os.environ.get('AWS_REGION_AIMS'))
      
      user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                        aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                        region_name=os.environ.get('AWS_REGION_AIMS'))
      
      s3 = user_sess.resource('s3')

      # Get bucket and model_id for user based on apiurl {{{
      response, error = run_function_on_lambda(
          apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
      )
      if error is not None:
          raise error

      _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
      # }}}

      email_list=json.loads(json.dumps(email_list))
    
       
      if update_type=="Replace_list":
          import json  
          import tempfile
          tempdir = tempfile.TemporaryDirectory()
          content_object = aws_client['resource'].Object(bucket_name=api_bucket, key=model_id +"/"+submission_type+"/competitionuserdata.json")
          file_content = content_object.get()['Body'].read().decode('utf-8')
          json_content = json.loads(file_content)
          json_content['emaillist']=email_list
          with open(tempdir.name+'/competitionuserdata.json', 'w', encoding='utf-8') as f:
              json.dump(json_content, f, ensure_ascii=False, indent=4)

          aws_client['client'].upload_file(
                tempdir.name+"/competitionuserdata.json", api_bucket, model_id +"/"+submission_type+"/competitionuserdata.json"
            )
      elif update_type=="Add":
          import json  
          import tempfile
          tempdir = tempfile.TemporaryDirectory()
          content_object = aws_client['resource'].Object(bucket_name=api_bucket, key=model_id +"/"+submission_type+"/competitionuserdata.json")
          file_content = content_object.get()['Body'].read().decode('utf-8')
          json_content = json.loads(file_content)

          email_list_old=json_content["emaillist"]
          email_list_new=email_list_old+email_list
          print(email_list_new)
            
          json_content["emaillist"]=email_list_new
          with open(tempdir.name+'/competitionuserdata.json', 'w', encoding='utf-8') as f:
              json.dump(json_content, f, ensure_ascii=False, indent=4)

          aws_client['client'].upload_file(
                tempdir.name+"/competitionuserdata.json", api_bucket, model_id +"/"+submission_type+"/competitionuserdata.json"
            )
     
          return "Success: Your competition participant access list is now updated."
      elif update_type=="Remove":
          import json  
          import tempfile
          tempdir = tempfile.TemporaryDirectory()
    
          aws_client['resource']
          content_object = aws_client['resource'].Object(bucket_name=api_bucket, key=model_id +"/"+submission_type+"/competitionuserdata.json")
          file_content = content_object.get()['Body'].read().decode('utf-8')
          json_content = json.loads(file_content)

          email_list_old=json_content["emaillist"]
          email_list_new=list(set(list(email_list_old)) - set(email_list))
          print(email_list_new)
            
          json_content["emaillist"]=email_list_new
          with open(tempdir.name+'/competitionuserdata.json', 'w', encoding='utf-8') as f:
              json.dump(json_content, f, ensure_ascii=False, indent=4)

          aws_client['client'].upload_file(
                tempdir.name+"/competitionuserdata.json", api_bucket, model_id +"/"+submission_type+"/competitionuserdata.json"
            )
          return "Success: Your competition participant access list is now updated."
      elif update_type=="Get":
          import json  
          import tempfile
          tempdir = tempfile.TemporaryDirectory()

          aws_client['resource']
          content_object = aws_client['resource'].Object(bucket_name=api_bucket, key=model_id +"/"+submission_type+"/competitionuserdata.json")
          file_content = content_object.get()['Body'].read().decode('utf-8')
          json_content = json.loads(file_content)
          return json_content['emaillist']
      else:
          return "Error: Check inputs and resubmit."


def _confirm_libraries_exist(requirements):
  requirements = requirements.split(",")
  for i in range(len(requirements)):
      requirements[i] = requirements[i].strip(" ")
      exists = requests.get("https://pypi.org/project/" + requirements[i]) 
      if exists.status_code == 404:
          try_again_message = ("The entered library '" + requirements[i] + "' was not found. "
                                "Please confirm and re-submit library name: ")
          requirements[i] = input(try_again_message)
          exists = requests.get("https://pypi.org/project/" + requirements[i])
  
          if exists.status_code == 404:
              error_message = ("ModuleNotFoundError: No module named '" + requirements[i] + "' found in the Python Package Index (PyPI). \n"
                               "Please double-check library name and try again.")
              return print(error_message)

  return



def _create_exampledata_json(model_type, exampledata_folder_filepath): 
    import tempfile
    # create temporary folder
    temp_dir = tempfile.gettempdir()

    image_extensions = ['.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.avif', 
                        '.svg', '.webp', '.tif', '.bmp', '.jpe', '.jif', '.jfif',
                        '.jfi', 'psd', '.raw', '.arw', '.cr2', '.nrw', '.k25', '.eps']
    video_extensions = ['.avchd', '.avi', '.flv', '.mov', '.mkv', '.mp4', '.wmv']
    audio_extensions = ['.m4a', '.flac', '.mp3', '.mp4', '.wav', '.wma', '.aac']

    import pandas as pd
     
    if any([model_type.lower() == "tabular", model_type.lower() == "timeseries", model_type.lower() == "text"]):
        #confirm data type is data frame, try to convert if not [necessary for front end]
        if isinstance(exampledata_folder_filepath, pd.DataFrame):
            pass
        else:
            exampledata_folder_filepath = pd.DataFrame(exampledata_folder_filepath)
            
        tabularjson = exampledata_folder_filepath.to_json(orient='split', index=False)
        
    
        with open(temp_dir+'/exampledata.json', 'w', encoding='utf-8') as f:
            json.dump({"exampledata": tabularjson, "totalfiles":1}, f, ensure_ascii=False, indent=4)

            return
        
    else:

        if isinstance(exampledata_folder_filepath, pd.DataFrame):

            tabularjson = exampledata_folder_filepath.to_json(orient='split', index=False)
            with open(temp_dir+'/exampledata.json', 'w', encoding='utf-8') as f:
                json.dump({"exampledata": tabularjson, "totalfiles":1}, f, ensure_ascii=False, indent=4)
        else:
            #Check file types & make list to convert 
            data = ""
            file_list = os.listdir(exampledata_folder_filepath)
            files_to_convert = []
            for i in range(len(file_list)):
                file_list[i] = exampledata_folder_filepath + "/" + file_list[i]
                root, ext = os.path.splitext(file_list[i])
                
                if not ext:
                    ext = mimetypes.guess_extension(file_list[i])
                    
                if model_type.lower() == "image" and ext in image_extensions:
                    files_to_convert.append(file_list[i])
                        
                if model_type.lower() == "video" and ext in video_extensions:
                    files_to_convert.append(file_list[i])
                    
                if model_type.lower() == "audio" and ext in audio_extensions: 
                    files_to_convert.append(file_list[i])     
            
                i += 1 
                if len(files_to_convert) == 5:
                    break
        
            #base64 encode confirmed file list 
            for i in range(len(files_to_convert)):
                with open(files_to_convert[i], "rb") as current_file: 
                    encoded_string = base64.b64encode(current_file.read())
                    data = data + encoded_string.decode('utf-8') + ", "
                    i += 1
        
            #build json
            with open(temp_dir+'/exampledata.json', 'w', encoding='utf-8') as f:
                json.dump({"exampledata": data[:-2], "totalfiles": len(files_to_convert)}, f, ensure_ascii=False, indent=4)
            
        return

__all__ = [
    take_user_info_and_generate_api,
    send_model_data_to_dyndb_and_return_api,
    model_to_api,
    create_competition,update_access_list
]
